use std::collections::HashSet;
use std::sync::Arc;
use tokio::sync::Mutex;

use serenity::{prelude::GatewayIntents, Client as SerenityClient};

use crate::{database::Client as DatabaseClient, environment::get_env_variable};

use crate::discord::event_handler::Handler;

/// The Yorokobot client.
///
/// To launch Yorokobot, you have to set the following environment variables:
/// - DISCORD_TOKEN: The secret Discord provide you when creating a new bot in the
/// Discord Developper websites.
/// - MONGODB_URI: The connection string to your Mongo database.
/// - MONGODB_DATABASE: The database to use in your Mongo instance (falcultative if given in the
/// MONGODB_URI connection string).
pub struct Client {
    serenity_client: SerenityClient,
}

impl Client {
    /// Create a new Yorokobot instance
    pub async fn new() -> Self {
        let mut database_client = DatabaseClient::new();

        database_client.connect().await;

        let discord_token = get_env_variable("DISCORD_TOKEN");
        let intents = GatewayIntents::GUILD_MESSAGES
            | GatewayIntents::MESSAGE_CONTENT
            | GatewayIntents::GUILD_MESSAGE_REACTIONS;

        let event_handler = Handler {
            database: Arc::new(database_client),
            users_with_running_selector: Arc::new(Mutex::new(HashSet::new())),
        };

        let serenity_client = match SerenityClient::builder(discord_token, intents)
            .event_handler(event_handler)
            .await
        {
            Ok(c) => c,
            Err(e) => panic!("Failed to instantiate Discord Client: {e}"),
        };

        Client { serenity_client }
    }

    /// Start the bot, connecting it to the database and the Discord API.
    pub async fn start(&mut self) {
        if let Err(e) = self.serenity_client.start().await {
            panic!("Could not connect the bot: {e}");
        }
    }
}
