//! The MongoDB crate wrapper used by Yorokobot to query the database.

use std::collections::HashSet;

use futures::TryStreamExt;
use log::{error, info, trace};
use mongodb::{
    bson::{doc, from_bson, to_document, Bson, Document},
    Client as MongoClient, Collection, Database,
};

use crate::environment::get_env_variable;

use super::models::{YorokobotCollection, COLLECTIONS_NAMES};

/// Yorokobot's database client.
///
/// Actually, it's just a wrapper around the [mongodb] crate.
/// It does basically nothing until you call the [`connect`] method.
///
/// To use it, the MONGODB_URI and/or MONGODB_DATABASE environment variable
/// have to be set.
///
/// [`connect`]: Client::connect
///
/// # Collections
///
/// To call the public method of this wrapper, you have to declare a Collection struct,
/// implement the [YorokobotCollection] traits, and add the collection name to [COLLECTIONS_NAMES].
/// See the [`Guild`] struct.
///
/// [`Guild`]: super::models::Guild
///
/// # Update methods
///
/// In each update method of the Client struct, the given update has to be nested in the `$set`
/// parameter, as this argument is passed as it is in the request to the database.
///
///# Example
///
/// ```
/// use mongodb::doc;
///
/// let mut db_client = Client::new();
///
/// db_client.connect();
///
/// // Do your stuff with the database.
/// let result = db_client.get_one::<Guild>(doc! {}).expect().unwrap();
///
/// db.update_one::<Guild>(result, doc! { "$set": { "tags": Vec::new() } });
/// ```
///
pub struct Client {
    mongo_client: Option<MongoClient>,
    database: Option<Database>,
}

impl Client {
    /// Create a new Database client instance.
    pub fn new() -> Client {
        Client {
            mongo_client: None,
            database: None,
        }
    }

    /// Get a reference to the inner mongodb client.
    fn get_database(&self) -> &Database {
        match &self.database {
            Some(db) => db,
            None => {
                error!("Tried to access to the database before instantiating it");
                panic!();
            }
        }
    }

    /// Connect the client.
    ///
    /// This method will panic if:
    /// - The MONGODB_URI environment variable is not set.
    /// - If the database to use is not specified in the connection string given with MONGODB_URI
    /// and if MONGODB_DATABASE is not set.
    /// - If the given credentials are invalid.
    pub async fn connect(&mut self) {
        self.mongo_client = match MongoClient::with_uri_str(get_env_variable("MONGODB_URI")).await {
            Ok(c) => {
                info!("Successfully connected to the Mongo database");
                Some(c)
            }
            Err(e) => {
                error!("Failed to connect to the Mongo database: {e:#?}");
                panic!();
            }
        };

        self.database = match &self.mongo_client {
            Some(c) => Some(c.database(get_env_variable("MONGODB_DATABASE").as_str())),
            None => {
                error!("Got an unexpected None for inner mongodb client instance");
                panic!();
            }
        };

        self.check_init_error().await;
    }

    /// Initials check that are run before right after connecting to the database.
    ///
    /// It currently runs the following tests:
    /// - Check that all the needed Mongo collections exists within the database.
    async fn check_init_error(&mut self) {
        info!("Launching initial database checks");

        self.check_collections_presence().await;
    }

    /// Check the collections presence within the database.
    ///
    /// This function is called by [`check_init_error`].
    /// This method panic if:
    /// - It fails to fetch the database collections.
    /// - A collection is missing.
    ///
    /// [`check_init_error`]: Self::check_init_error
    async fn check_collections_presence(&self) {
        trace!("Starting the collections presence check for the database");

        let mut missing_collections: Vec<&str> = vec![];
        let collections: HashSet<String> =
            match self.get_database().list_collection_names(None).await {
                Ok(n) => n.into_iter().collect(),
                Err(e) => {
                    error!("Failed to get the collections for the database: {e:#?}");
                    panic!();
                }
            };

        for col in COLLECTIONS_NAMES {
            if !collections.contains(col) {
                missing_collections.push(col);
            }
        }

        if !missing_collections.is_empty() {
            error!(
                "Missing the following collections in the Database: {}",
                missing_collections.join(", ")
            );
            panic!();
        }
    }

    /// Get access to the collection matching the `T` type.
    ///
    /// Unlike the [`get_typed_collection`] method, this one allows to deals directly with mongodb
    /// [Document] type, without using [serde] (de)serialization.
    ///
    /// [`get_typed_collection`]: Self::get_typed_collection
    #[allow(dead_code)]
    fn get_collection<T: YorokobotCollection>(&self) -> Collection<Document> {
        self.get_database().collection(&T::get_collection_name())
    }

    /// Get access to the collection matching the `T` type, but dealing directly with the matching
    /// struct.
    ///
    /// This internly use [serde]'s [`Serialize`] and [`Deserialize`] traits.
    ///
    /// [`Serialize`]: serde::Serialize
    /// [`Deserialize`]: serde::Deserialize
    fn get_typed_collection<T: YorokobotCollection>(&self) -> Collection<T> {
        self.get_database()
            .collection::<T>(&T::get_collection_name())
    }

    /// Return the document in the `T` type matching the `id` ObjectId.
    ///
    /// Return Ok(None) if the collection does not contains element matching this id and Err(())
    /// if the query failed.
    pub async fn get_by_id<T: YorokobotCollection>(&self, id: &str) -> Result<Option<T>, ()> {
        self.get_one::<T>(doc! {"_id": id}).await
    }

    /// Return the first element of the `T` type matching the given `filter`.
    ///
    /// Return Ok(None) if the collection does not contains matching element matching this filter
    /// and Err(()) if the query failed.
    pub async fn get_one<T: YorokobotCollection>(&self, filter: Document) -> Result<Option<T>, ()> {
        match self
            .get_typed_collection::<T>()
            .find_one(filter, None)
            .await
        {
            Ok(e) => Ok(e),
            Err(_) => Err(()),
        }
    }

    /// Return all the element of the type `T`.
    /// If `filter` is provided, return only the ones matching the given filter.
    #[allow(dead_code)]
    pub async fn get_all<T: YorokobotCollection>(
        &self,
        filter: Option<Document>,
    ) -> Result<Vec<T>, ()> {
        let mut matching_docs: Vec<T> = vec![];

        let result = match filter {
            Some(f) => self.get_collection::<T>().find(f, None).await,
            None => self.get_collection::<T>().find(doc! {}, None).await,
        };

        let mut cursor = match result {
            Ok(c) => c,
            Err(_) => return Err(()),
        };

        while let Some(document) = match cursor.try_next().await {
            Ok(e) => e,
            Err(_) => return Err(()),
        } {
            match from_bson(Bson::Document(document)) {
                Ok(d) => matching_docs.push(d),
                Err(_) => return Err(()),
            }
        }

        Ok(matching_docs)
    }

    /// Insert the provided `document` in the database.
    /// Return Err(()) if the query failed.
    pub async fn insert_one<T: YorokobotCollection>(&self, document: T) -> Result<(), ()> {
        match self
            .get_typed_collection::<T>()
            .insert_one(document, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }

    /// Insert all the provided `documents` in the database.
    /// Return Err(()) if the query failed.
    #[allow(dead_code)]
    pub async fn insert_many<T: YorokobotCollection>(&self, documents: Vec<T>) -> Result<(), ()> {
        match self
            .get_typed_collection::<T>()
            .insert_many(documents, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }

    /// Delete the first document matching the given `document`.
    /// Return Err(()) if the query failed.
    #[allow(dead_code)]
    pub async fn delete_one<T: YorokobotCollection>(&self, document: Document) -> Result<u64, ()> {
        match self
            .get_typed_collection::<T>()
            .delete_one(document, None)
            .await
        {
            Ok(result) => Ok(result.deleted_count),
            Err(_) => Err(()),
        }
    }

    /// Delete the document matching whose ObjectId match the given `id`.
    /// Return Err(()) if the query failed.
    #[allow(dead_code)]
    pub async fn delete_by_id<T: YorokobotCollection>(&self, id: &str) -> Result<(), ()> {
        match self
            .get_typed_collection::<T>()
            .delete_one(doc! {"_id": id}, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }

    /// Delete the documents matching the given filter.
    /// Return Err(()) if the query failed.
    #[allow(dead_code)]
    pub async fn delete_many<T: YorokobotCollection>(&self, filter: Document) -> Result<(), ()> {
        match self
            .get_typed_collection::<T>()
            .delete_many(filter, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }

    /// Update the given `object` with the `update` modifications.
    /// Currently, we should pass the `"$set"` argument to the `update`
    /// document which is passed as it is to MongoDB.
    /// Return Err(()) if the query failed.
    pub async fn update_one<T: YorokobotCollection>(
        &self,
        object: T,
        update: Document,
    ) -> Result<(), ()> {
        let serialized_doc = match to_document(&object) {
            Ok(d) => d,
            Err(_) => return Err(()),
        };

        match self
            .get_typed_collection::<T>()
            .update_one(serialized_doc, update, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }

    /// Update the object whose ObjectId match the given `id` with the given `update`. The `update`
    /// parameter need to be nested within the MongoDB `"$set"` parameter, as it is passed as it is
    /// to the database instance.
    /// Return Err(()) if the query failed.
    #[allow(dead_code)]
    pub async fn update_by_id<T: YorokobotCollection>(
        &self,
        document_id: &str,
        update: Document,
    ) -> Result<(), ()> {
        match self
            .get_typed_collection::<T>()
            .update_one(doc! {"_id": document_id}, update, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }

    /// Update all the objects matching the given `filter` with the given `update`.
    /// The update parameter need to be nested within the MongoDB `"$set"` parameter, as it is
    /// passed as it is in the request.
    #[allow(dead_code)]
    pub async fn update_many<T: YorokobotCollection>(
        &self,
        filter: Document,
        update: Document,
    ) -> Result<(), ()> {
        match self
            .get_typed_collection::<T>()
            .update_many(filter, update, None)
            .await
        {
            Ok(_) => Ok(()),
            Err(_) => Err(()),
        }
    }
}
