//! Commons elements that are used in while executing the bot commands.

use log::{info, warn};
use std::sync::Arc;

use serenity::{
    async_trait,
    model::prelude::{
        command::{Command, CommandOptionType},
        interaction::application_command::{
            ApplicationCommandInteraction, CommandDataOption, CommandDataOptionValue,
        },
    },
    prelude::Context,
};

use crate::database::Client as DatabaseClient;

/// The kind of errors that can be returned while executing a command.
#[derive(Debug)]
pub enum CommandExecutionError {
    /// An error that is returned when the `Serenity` crate failed to get the argument from the
    /// Discord API payload.
    ArgumentExtractionError(String),

    /// An error that is returned when the cast from the argument value to a Rust type failed.
    ArgumentDeserializationError(String),

    /// The kind of error that is returned when a Database query failed.
    DatabaseQueryError(String),

    /// Error returned when we fail to extract the context of a Discord commands.
    ContextRetrievalError(String),

    /// Error returned when sending a command to the discord API failed.
    DiscordAPICallError(String),

    /// Error returned when there was an issue while using a selector embed in the command
    /// response.
    SelectorError(String),

    /// Error returned when the given command is unknown to Yorokobot.
    UnknownCommand(String),
}

/// An option that can/have to be passed to a Discord command.
pub struct BotCommandOption {
    pub name: String,
    pub description: String,
    pub kind: CommandOptionType,
    pub required: bool,
}

/// A trait that represent a Yorokobot command.
#[async_trait]
pub trait BotCommand {
    /// Create a new command instance.
    fn new(interaction: ApplicationCommandInteraction) -> Self;

    /// Return the name of the command.
    fn name() -> String;

    /// Return the command description.
    fn description() -> String;

    /// Return the list of the argument that the command takes.
    fn options_list() -> Vec<BotCommandOption>;

    /// Execute the command with the given `context` and using the given `database`.
    async fn run(
        &self,
        context: Context,
        database: Arc<DatabaseClient>,
    ) -> Result<(), CommandExecutionError>;

    /// Extract and deserialize the `index`th value from the command `options`.
    fn extract_option_value(
        options: &[CommandDataOption],
        index: usize,
    ) -> Option<CommandDataOptionValue> {
        let serialized_opt = options.get(index)?;
        serialized_opt.resolved.clone()
    }

    /// Register the command to the Discord API.
    async fn register(context: &Context) {
        info!("Starting the {} command registration", Self::name());
        match Command::create_global_application_command(context, |command| {
            let mut new_command = command.name(Self::name()).description(Self::description());

            for opt in Self::options_list() {
                new_command = new_command.create_option(|option| {
                    option
                        .name(opt.name)
                        .description(opt.description)
                        .kind(opt.kind)
                        .required(opt.required)
                });
            }
            new_command
        })
        .await
        {
            Ok(_) => info!("Successfully registered the {} command", Self::name()),
            Err(_) => warn!("Failed to register the {} command", Self::name()),
        };
    }
}
