use std::sync::Arc;

use log::debug;

use serenity::{
    async_trait,
    model::prelude::interaction::{
        application_command::ApplicationCommandInteraction, InteractionResponseType,
    },
    prelude::Context,
};

use crate::database::{models::Guild, Client as DatabaseClient};

use super::{
    commons::{BotCommandOption, CommandExecutionError},
    BotCommand,
};

pub struct ListTagCommand {
    interaction: ApplicationCommandInteraction,
}

#[async_trait]
impl BotCommand for ListTagCommand {
    fn name() -> String {
        String::from("list_tags")
    }

    fn description() -> String {
        String::from("List available tags")
    }

    fn options_list() -> Vec<BotCommandOption> {
        vec![]
    }

    fn new(interaction: ApplicationCommandInteraction) -> Self {
        debug!("Creating a new ListTagCommand object");
        ListTagCommand { interaction }
    }

    async fn run(
        &self,
        context: Context,
        database: Arc<DatabaseClient>,
    ) -> Result<(), CommandExecutionError> {
        let guild_id = match self.interaction.guild_id {
            Some(id) => Ok(id.to_string()),
            None => Err(CommandExecutionError::ContextRetrievalError(
                "Failed to extract guild id from current context".to_string(),
            )),
        }?;

        let guild = match database.get_by_id::<Guild>(&guild_id).await {
            Ok(query) => match query {
                Some(r) => Ok(r),
                None => Err(CommandExecutionError::ContextRetrievalError(
                    "Failed to retrieve the guild where the command was issued".to_string(),
                )),
            },
            Err(()) => Err(CommandExecutionError::DatabaseQueryError(
                "Failed to access to the database".to_string(),
            )),
        }?;

        let mut response_content: String;

        if guild.tags.is_empty() {
            response_content = String::from("No tag available on this server.");
        } else {
            response_content = String::from("Available tags on this server:\n");
            for tag in guild.tags {
                if response_content.len() + tag.name.len() < 1995 {
                    response_content.push_str(format!("`{}` ", tag.name.as_str()).as_str());
                } else {
                    response_content.push_str("...");
                }
            }
        }

        match self
            .interaction
            .create_interaction_response(context.http, |response| {
                response
                    .kind(InteractionResponseType::ChannelMessageWithSource)
                    .interaction_response_data(|message| message.content(response_content))
            })
            .await
        {
            Ok(()) => Ok(()),
            Err(_) => Err(CommandExecutionError::DiscordAPICallError(
                "Failed to answer to the initial command".to_string(),
            )),
        }
    }
}
